/*
 * Set the framebuffer parameters for systems without DGA enabled X server.
 *   tries to check /dev/fb0 or one defined in FRAMEBUFFER
 *
 *  This is a stripped down version of v4l-conf.c of xawtv-2.36 by
 *  (c) 1998 Gerd Knorr <kraxel@goldbach.in-berlin.de>, and any problem
 *  you might encounter is not Gerd's fault.  Please send a bug report to
 *  me.
 *
 *  Takashi Oe <toe@unlinfo.unl.edu>
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <sys/ioctl.h>

#include <linux/vt.h>
#include <linux/fb.h>
#include <linux/videodev.h>

//#include "fb.h"
#include "vc_ioctl.h"
#include "planb.h"

int    verbose = 1;
int    bpp     = 0;
char  *device  = "/dev/video";
char  *fbdev = NULL;
int    set_fb  = 0;
int    regs    = 0;
int    sregs   = 0;
int    mode    = 0;
int    peek    = 0;

int is_mklinux() {
	int fd;
	if(-1 == (fd = open("/proc/osfmach3", O_RDONLY)))
		return 0;
	close(fd);
	return 1;
}

int
main(int argc, char *argv[])
{
    /* fbdev */
    struct fb_fix_screeninfo   fix;
    struct fb_var_screeninfo   var;
    struct fb_con2fbmap        c2m;
    
    /* v4l */
    struct video_capability  capability;
    struct video_buffer      fbuf;
    int                      set_width=0, set_height=0, set_bpp=0, set_bpl=0;
    void                     *set_base=NULL;

    /* misc */
    int                      fd,c,i,n;
    char                     *h;
    extern int               optind;

    
    /* take defaults from environment */
    if (NULL != (h = getenv("FRAMEBUFFER")))
	fbdev = h;
    
    /* parse options */
    for (;;) {
	if (-1 == (c = getopt(argc, argv, "hqf:c:ripstv")))
	    break;
	switch (c) {
	case 'q':
	    verbose = 0;
	    break;
	case 'f':
	    fbdev = optarg;
	    break;
	case 'c':
	    device = optarg;
	    break;
	case 'r':
	    regs = 1;
	    break;
	case 'i':
	    peek = 1;
	    break;
	case 'p':
	    sregs = 1;
	    break;
	case 's':
	    set_fb = 1;
	    break;
	case 't':
	    mode = PLANB_TV_MODE;
	    break;
	case 'v':
	    mode = PLANB_VTR_MODE;
	    break;
	case 'h':
	default:
	    fprintf(stderr,
		    "usage: %s  [ options ] [addr val]\n"
		    "\n"
		    "options:\n"
		    "    -q        quiet\n"
		    "    -r        manipulate saa regs\n"
		    "    -p        show planb status\n"
		    "    -t        set planb to TV mode\n"
		    "    -v        set planb to VTR mode\n"
		    "    -s        set frame buffer address\n"
		    "    -f <dev>  frame buffer device [%s]\n"
		    "    -c <dev>  video device        [%s]\n",
		    argv[0],
		    fbdev ? fbdev : "none",
		    device);
	    exit(1);
	}
    }

    if(set_fb) {
	if (is_mklinux()) {
	    struct vc_mode mode;

	    if (-1 == (fd = open("/dev/console",O_RDWR|O_NDELAY))) {
		fprintf(stderr,"open console: %s\n",strerror(errno));
		exit(1);
	    }
	    if (-1 == ioctl(fd, VC_GETMODE, (unsigned long)&mode)) {
		perror("ioctl VC_GETMODE");
		exit(1);
	    }
	    close(fd);
	    set_width  = mode.width;
	    set_height = mode.height;
	    set_bpp    = mode.depth;
	    set_bpl    = mode.pitch;
	    set_base   = (void*)mode.fb_address;
	} else {
	    /* try framebuffer device */
	    if (NULL == fbdev) {
		struct vt_stat vstat;

		if (-1 == (fd = open("/dev/tty0",O_RDWR,0))) {
		    fprintf(stderr,"open /dev/tty0: %s\n",strerror(errno));
		    exit(1);
		}
		if (-1 == ioctl(fd, VT_GETSTATE, &vstat)) {
		    perror("ioctl VT_GETSTATE");
		    exit(1);
		}
		close(fd);
		c2m.console = vstat.v_active;
		if (-1 == (fd = open("/dev/fb0",O_RDWR,0))) {
		    fprintf(stderr,"open /dev/fb0: %s\n",strerror(errno));
		    goto bailout2;
		}
		if (-1 == ioctl(fd, FBIOGET_CON2FBMAP, &c2m)) {
		    perror("ioctl FBIOGET_CON2FBMAP");
		    goto bailout1;
		}
		close(fd);
		fprintf(stderr,"map: vt%02d => fb%d\n", c2m.console,
							c2m.framebuffer);
		sprintf(fbdev=malloc(16),"/dev/fb%d",c2m.framebuffer);
	    }
	    if (-1 == (fd = open(fbdev,O_RDWR))) {
		fprintf(stderr,"open %s: %s\n",fbdev,strerror(errno));
		goto bailout2;
	    }
	    if (-1 == ioctl(fd,FBIOGET_FSCREENINFO,&fix)) {
		perror("ioctl FBIOGET_FSCREENINFO");
		goto bailout1;
	    }
	    if (-1 == ioctl(fd,FBIOGET_VSCREENINFO,&var)) {
		perror("ioctl FBIOGET_VSCREENINFO");
		goto bailout1;
	    }
	    if (fix.type != FB_TYPE_PACKED_PIXELS) {
		fprintf(stderr,"can handle only packed pixel frame buffers\n");
		goto bailout1;
	    }
	    close(fd);
	    if (verbose)
		fprintf(stderr,"using framebuffer device %s: %s\n",fbdev,
									fix.id);

	    set_width  = var.xres_virtual;
	    set_height = var.yres_virtual;
	    set_bpp    = var.bits_per_pixel;
	    set_bpl    = fix.line_length;
	    set_base   = fix.smem_start;
	    goto success1;
bailout1:
	    close(fd);
bailout2:
	    set_fb = 0;
success1:
	}
    }

    /* open & check v4l device */
    if (-1 == (fd = open(device,O_RDWR))) {
	fprintf(stderr,"can't open %s: %s\n",device,strerror(errno));
	exit(1);
    }
    if (-1 == ioctl(fd,VIDIOCGCAP,&capability)) {
	fprintf(stderr,"%s: ioctl VIDIOCGCAP: %s\n",device,strerror(errno));
	goto bailout3;
    }
    if (strncmp(capability.name, PLANB_DEVICE_NAME, strlen(PLANB_DEVICE_NAME))){
	fprintf(stderr,"%s: not PlanB\n",device);
	goto bailout3;
    }
    if(set_fb) {
	if (!(capability.type & VID_TYPE_OVERLAY)) {
	    fprintf(stderr,"%s: no overlay support\n",device);
	    goto bailout4;
	}
	/* read-modify-write v4l screen parameters */
	if (-1 == ioctl(fd,VIDIOCGFBUF,&fbuf)) {
	    fprintf(stderr,"%s: ioctl VIDIOCGFBUF: %s\n",device,
							strerror(errno));
	    goto bailout4;
	}
	/* set values */
	fbuf.width        = set_width;
	fbuf.height       = set_height;
	fbuf.depth        = set_bpp;
	fbuf.bytesperline = set_bpl;
	fbuf.base     = set_base;
	if (verbose) {
	    fprintf(stderr,"video mode: %dx%d, %d bit/pixel, %d "
		"byte/scanline\n", fbuf.width,fbuf.height,fbuf.depth,
		fbuf.bytesperline);
	    fprintf(stderr,"framebuffer at %p\n",set_base);
	}
	if (-1 == ioctl(fd,VIDIOCSFBUF,&fbuf)) {
	    fprintf(stderr,"%s: ioctl VIDIOCSFBUF: %s\n",device,
							strerror(errno));
	    goto bailout4;
	}
bailout4:
	set_fb = 0;
    }
    if(mode) {
	int pmode;

	if (-1 == ioctl(fd,PLANBIOCGMODE,&pmode)) {
	    fprintf(stderr,"%s: ioctl PLANBIOCGMODE: %s\n", device,
							strerror(errno));
	    goto bailout3;
	}
	if (pmode != mode) {
	    if (-1 == ioctl(fd,PLANBIOCSMODE,&mode)) {
		fprintf(stderr,"%s: ioctl PLANBIOCSMODE: %s\n", device,
							strerror(errno));
		goto bailout3;
	    }
	    if (verbose)
		fprintf(stderr, "changed to %s mode\n", (mode == PLANB_TV_MODE)
							? "TV": "VTR");
	} else if (verbose)
	    fprintf(stderr, "mode unchanged\n");
    }
    if(regs || peek) {
	struct planb_any_regs any;
	struct planb_saa_regs preg;

	if(argv[optind]) {
	    char *endptr = argv[optind];

	    if(!(peek))
		preg.addr = (unsigned char)strtoul(argv[optind], &endptr, 0);
	    else
		any.offset = (unsigned int)strtoul(argv[optind], &endptr, 0);
	    if(endptr != argv[optind]) {
		if(argv[optind + 1]) {
		    endptr = argv[optind + 1];
		    if(!(peek))
			preg.val = (unsigned char)strtoul(argv[optind + 1],
								&endptr, 0);
		    else {
			any.bytes = (unsigned int)strtoul(argv[optind + 1],
								&endptr, 0);
			any.bytes += any.bytes % 4;
		    }
		    if(endptr != argv[optind + 1]) {
			if(!(peek)) {
			    if (-1 == ioctl(fd,PLANBIOCSSAAREGS,&preg)) {
				fprintf(stderr,"%s: ioctl PLANBIOCSSAAREGS: "
						"%s\n", device,strerror(errno));
				goto bailout3;
			    }
			} else {
			    if (-1 == ioctl(fd, PLANB_INV_REGS,&any)) {
				fprintf(stderr,"%s: ioctl PLANBIOCSSAAREGS: "
						"%s\n", device,strerror(errno));
				goto bailout3;
			    }
			}
		    }
		} else if (regs) {
		    if (-1 == ioctl(fd,PLANBIOCGSAAREGS,&preg)) {
			fprintf(stderr,"%s: ioctl PLANBIOCGSAAREGS: %s\n",
							device,strerror(errno));
			goto bailout3;
		    }
		}
		if(!(peek))
		    printf("addr=0x%x, val=0x%x\n", preg.addr, preg.val);
		else {
		    printf("addr=0x%x\n", any.offset + 0xf1000000);
		    for(i = 0; i < any.bytes; i+=4) {
			int j;

			if(!(i % 16) && (i))
			    printf("\n");
			printf(" ");
			for(j = 0; j < 4; j++)
			    printf("%02X", any.data[i+j]);
		    }
		    printf("\n");
		}
	    }
	}
    }
    if(sregs) {
	struct planb_stat_regs pstat;

	if (-1 == ioctl(fd,PLANBIOCGSTAT,&pstat)) {
	    fprintf(stderr,"%s: ioctl PLANBIOCGSTAT: %s\n", device,
							strerror(errno));
	    goto bailout3;
	}
	printf("ch1=0x%x, ch2=0x%x, saa0=0x%x, saa1=0x%x\n", pstat.ch1_stat,
		pstat.ch2_stat, pstat.saa_stat0, pstat.saa_stat1);
    }
    close(fd);
    if (verbose)
	fprintf(stderr,"ok\n");
    return 0;

bailout3:
    close(fd);
    return 1;
}

